/*<std-header orig-src='shore' incl-file-exclusion='KVL_T_H'>

 $Id: kvl_t.h,v 1.12 2012/01/02 21:52:20 nhall Exp $

SHORE -- Scalable Heterogeneous Object REpository

Copyright (c) 1994-99 Computer Sciences Department, University of
                      Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE AUTHORS AND THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY
OF WISCONSIN - MADISON ALLOW FREE USE OF THIS SOFTWARE IN ITS
"AS IS" CONDITION, AND THEY DISCLAIM ANY LIABILITY OF ANY KIND
FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.

This software was developed with support by the Advanced Research
Project Agency, ARPA order number 018 (formerly 8230), monitored by
the U.S. Army Research Laboratory under contract DAAB07-91-C-Q518.
Further funding for this work was provided by DARPA through
Rome Research Laboratory Contract No. F30602-97-2-0247.

*/

#ifndef KVL_T_H
#define KVL_T_H

#include "w_defines.h"

/*  -- do not edit anything above this line --   </std-header>*/

/* NB: you must already have defined the type size_t,
 * (which is defined include "basics.h") before you include this.
 */

#ifdef __GNUG__
#pragma interface
#endif

#ifndef STID_T_H
#include "stid_t.h"
#endif
#ifndef VEC_T_H
#include "vec_t.h"
#endif

/**\brief Key-Value Lock identifier.
 *
 * \details
 * Used by the lock manager.
 */
struct kvl_t {
    stid_t            stid;
    w_base_t::uint4_t        h;
    w_base_t::uint4_t        g;

    static const cvec_t eof;
    static const cvec_t bof;

    // Empty key-value
    NORET            kvl_t();
    // Empty key-value lock id for a store ID.
    NORET            kvl_t(stid_t id, const cvec_t& v);
    // Empty key-value lock id for a store ID.
    NORET            kvl_t(
        stid_t                _stid,
        const cvec_t&         v1, 
        const cvec_t&         v2);

    NORET            ~kvl_t();
    
    NORET            kvl_t(const kvl_t& k);

    kvl_t&             operator=(const kvl_t& k);

    kvl_t&             set(stid_t s, const cvec_t& v);
    kvl_t&             set(stid_t s,
        const cvec_t&         v1,
        const cvec_t&         v2);
    bool operator==(const kvl_t& k) const;
    bool operator!=(const kvl_t& k) const;
    friend ostream& operator<<(ostream&, const kvl_t& k);
    friend istream& operator>>(istream&, kvl_t& k);
};

inline NORET
kvl_t::kvl_t()
    : stid(stid_t::null), h(0), g(0)
{
}

inline NORET
kvl_t::kvl_t(stid_t id, const cvec_t& v)
    : stid(id)
{
	// don't waste "g" for uni_btree. uses different hash seed
    v.calc_kvl(h), v.calc_kvl2(g); 
    v.calc_kvl(h), v.calc_kvl2(g); 
}

inline NORET
kvl_t::kvl_t(stid_t id, const cvec_t& v1, const cvec_t& v2)
    : stid(id)  
{
    v1.calc_kvl(h); v2.calc_kvl(g);
}

inline NORET
kvl_t::~kvl_t()
{
}

inline NORET
kvl_t::kvl_t(const kvl_t& k)
    : stid(k.stid), h(k.h), g(k.g)
{
}

inline kvl_t& 
kvl_t::operator=(const kvl_t& k)
{
    stid = k.stid;
    h = k.h, g = k.g;
    return *this;
}
    

inline kvl_t&
kvl_t::set(stid_t s, const cvec_t& v)
{
	// don't waste "g" for uni_btree. uses different hash seed
    stid = s, v.calc_kvl(h), v.calc_kvl2(g); 
    return *this;
}

inline kvl_t& 
kvl_t::set(stid_t s, const cvec_t& v1, const cvec_t& v2)
{
    stid = s, v1.calc_kvl(h), v2.calc_kvl(g);
    return *this;
}

inline bool
kvl_t::operator==(const kvl_t& k) const
{
    return h == k.h && g == k.g;
}

inline bool
kvl_t::operator!=(const kvl_t& k) const
{
    return ! (*this == k);
}

/*<std-footer incl-file-exclusion='KVL_T_H'>  -- do not edit anything below this line -- */

#endif          /*</std-footer>*/
